@@ -40,16 +40,25 @@ struct arglist  script_dependencie(harglst*, struct arglist *);
 struct arglist	script_require_keys(harglst*, struct arglist *);
 struct arglist  script_exclude_keys(harglst*, struct arglist *);
 struct arglist  script_require_ports(harglst*, struct arglist *);
+struct arglist  script_require_udp_ports(harglst*, struct arglist *);
 struct arglist  script_add_preference(harglst*, struct arglist *);
 struct arglist  script_get_preference(harglst*, struct arglist *);
 struct arglist  script_timeout(harglst* , struct arglist *);
+struct arglist  script_id(harglst *, struct arglist *);
+struct arglist  script_cve_id(harglst *, struct arglist *);
+struct arglist  script_see_also(harglst *, struct arglist *);
 
 struct arglist  get_hostname(harglst*, struct arglist *);
 struct arglist  get_host_ip(harglst*, struct arglist *);
 struct arglist  get_host_open_port(harglst*, struct arglist *);
 struct arglist  get_port_state(harglst*, struct arglist *);
+struct arglist  get_udp_port_state(harglst*, struct arglist *);
 
 struct arglist pkt_open_sock_tcp(harglst*, struct arglist *);
+
+struct arglist pkt_open_priv_sock_tcp(harglst*, struct arglist*);
+struct arglist pkt_open_priv_sock_udp(harglst*, struct arglist*);
+
 struct arglist pkt_ftp_log_in(harglst*, struct arglist *);
 struct arglist pkt_ftp_get_pasv_address(harglst*, struct arglist *);
 struct arglist pkt_is_cgi_installed(harglst*, struct arglist *);
@@ -68,6 +77,8 @@ struct arglist pkt_send(harglst*, struct arglist *);
 struct arglist telnet_init(harglst*, struct arglist *);
 
 
+
+
 struct arglist  script_timeout(globals, vars)
  harglst * globals;
  struct arglist * vars;
@@ -88,6 +99,51 @@ struct arglist  script_timeout(globals, vars)
 }
 
 
+struct arglist  script_id(globals, vars)
+ harglst * globals;
+ struct arglist * vars;
+{
+ struct arglist * script_infos = harg_get_ptr(globals, "script_infos");
+ struct arglist rt;
+ bzero(&rt, sizeof(rt));
+ if(!vars->value)return rt;
+ rt = sanitize_variable(globals, vars->value);
+ if(rt.type)
+ {
+  plug_set_id(script_infos, atoi(rt.value));
+ }
+ bzero(&rt, sizeof(rt));
+ return rt;
+}
+
+struct arglist  script_cve_id(globals, vars)
+ harglst * globals;
+ struct arglist * vars;
+{
+ struct arglist * script_infos = harg_get_ptr(globals, "script_infos");
+ struct arglist rt;
+ bzero(&rt, sizeof(rt));
+ if(!vars->value)return rt;
+ rt = sanitize_variable(globals, vars->value);
+ if(rt.type)
+ {
+  plug_set_cve_id(script_infos, rt.value);
+ }
+ bzero(&rt, sizeof(rt));
+ return rt;
+}
+
+struct arglist  script_see_also(globals, vars)
+ harglst * globals;
+ struct arglist * vars;
+{
+ struct arglist rt;
+ bzero(&rt, sizeof(rt));
+ fprintf(stderr, "Error - script_see_also() called\n");
+ return rt;
+}
+
+
 static struct arglist script_elem(globals, vars, name)
  harglst * globals;
  struct arglist * vars;
@@ -287,6 +343,36 @@ struct arglist  script_require_ports(globals, vars)
  
  return(rt);
 }
+
+
+struct arglist  script_require_udp_ports(globals, vars)
+ harglst * globals;
+ struct arglist * vars;
+{
+ struct arglist rt;
+ struct arglist * script_infos = harg_get_ptr(globals, "script_infos");
+ struct arglist sa;
+ 
+ if(!vars->value){
+ 	printf("Argument error in function script_require_udp_ports()\n");
+	printf("Function usage is : script_require_ports(<name>)\n");
+	printf("Where <name> is the number of a port\n");
+	rt.type = PKT_ERROR(ERR_FUNC|ERR_ARG);
+	return(rt);
+	}
+  
+ bzero(&rt, sizeof(rt));
+ while(vars && vars->next)
+ {
+ sa = sanitize_variable(globals, vars->value);
+ if(sa.type & VAR_STR){
+   	 plug_require_udp_port(script_infos, sa.value);
+	}
+ if(sa.type & VAR_DELETE)nasl_free(globals, sa.value);
+ vars = vars->next;
+ }
+ return(rt);
+}
 struct arglist  script_add_preference(globals, args)
  harglst * globals;
  struct arglist * args;
@@ -434,6 +520,41 @@ get_port_state(globals, vars)
  return(rt);
 }
 
+
+struct arglist
+get_udp_port_state(globals, vars)
+ harglst * globals;
+ struct arglist *vars;
+{
+ int open;
+ struct arglist rt;
+ struct arglist *  script_infos = harg_get_ptr(globals, "script_infos");
+ char * value = nasl_malloc(globals,2);
+ 
+ bzero(&rt, sizeof(rt));
+ value[0]='0';
+ rt.value = value;
+ rt.length = 0;
+ rt.type = VAR_STR|STR_ALL_DIGIT;
+ if(vars->value)
+ {
+  struct arglist sa;
+  sa = sanitize_variable(globals, vars->value);
+  if(sa.type)
+  {
+   open = host_get_port_state_udp(script_infos, atoi(sa.value));
+#ifdef DEBUG_NASL
+   value[0]='1';
+#else
+   if(open)value[0]='1';
+#endif   
+   rt.value = value;
+   rt.length = 1;
+  }
+ }
+ return(rt);
+}
+
 struct arglist
 pkt_close(globals, args)
  harglst * globals;
@@ -454,6 +575,129 @@ pkt_close(globals, args)
  return(rt);
 }
 
+static struct arglist
+pkt_open_priv_sock(globals, vars, proto)
+ harglst * globals;
+ struct arglist * vars;
+ int proto;
+{
+ struct arglist rt;
+ struct arglist * script_infos = harg_get_ptr(globals, "script_infos");
+ char * t;
+ int sport, current_sport = -1;
+ int dport;
+ int sock;
+ int e;
+ struct sockaddr_in addr, daddr;
+ struct in_addr * p;
+ 
+ bzero(&rt, sizeof(rt));
+ 
+ t = arg_get_value(vars, "sport");
+ if(t)sport = atoi(t);
+ else sport = -1;
+ 
+ t = arg_get_value(vars, "dport");
+ if(t)dport = atoi(t);
+ else
+  {
+   fprintf(stderr, "nasl : error in open_priv_sock_tcp()\n");
+   return rt;
+  }
+ 
+ if(proto == IPPROTO_TCP)
+  sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
+ else
+  sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
+  
+ 
+ if(sock < 0)return rt;
+ bzero(&addr,sizeof(addr));
+ if(sport > 0)
+  addr.sin_port = htons(sport);
+ else 
+  {
+  current_sport = 1023;
+  addr.sin_port = htons(current_sport);
+  addr.sin_addr.s_addr = INADDR_ANY;
+  addr.sin_family = AF_INET;
+  }
+tryagain :
+ e =  bind(sock, (struct sockaddr*)&addr, sizeof(addr));
+ /*
+  * try to bind on another priv port
+  */
+ if(e < 0)
+ {
+  if(sport > 0)
+  {
+   close(sock);
+   return rt;
+  }
+  else
+  {
+   current_sport --;
+   if(!current_sport)
+   {
+    close(sock);
+    return rt;
+   }
+   else {
+    addr.sin_port = htons(current_sport);
+    goto tryagain;
+    }
+  }
+ }
+ 
+ /*
+  * We are binded to a privileged port now. Let's declare
+  * it ready for reuse
+  */
+ e = 1;
+ setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &e, sizeof(int));
+ 
+ /*
+  * Connect to the other end
+  */
+ p = plug_get_host_ip(script_infos);
+ bzero(&daddr, sizeof(daddr));
+ daddr.sin_addr = *p;
+ daddr.sin_port = htons(dport);
+ daddr.sin_family = AF_INET;
+ 
+ e = connect(sock, (struct sockaddr*)&daddr, sizeof(daddr));
+ if(e < 0)
+ {
+  close(sock);
+  return rt;
+ }
+ else
+ {
+  rt.value = (void*)sock;
+  rt.type = VAR_INT;
+  rt.length = sizeof(int);
+  return rt;
+ }
+}
+
+struct arglist
+pkt_open_priv_sock_tcp(globals, vars)
+ harglst * globals;
+ struct arglist * vars;
+{
+ return pkt_open_priv_sock(globals, vars, IPPROTO_TCP);
+}
+
+struct arglist
+pkt_open_priv_sock_udp(globals, vars)
+ harglst * globals;
+ struct arglist * vars;
+{
+ return pkt_open_priv_sock(globals, vars, IPPROTO_UDP);
+}
+ 
+ 
+ 
 struct arglist
 pkt_open_sock_tcp(globals, vars)
  harglst * globals;
@@ -481,8 +725,11 @@ pkt_open_sock_tcp(globals, vars)
   rt.value = (void*)soc;
   rt.type  = VAR_INT;
   rt.length = sizeof(int);
-  if(soc < 0){	rt.value = 0;
+  if(soc < 0){	
+  		rt.value = 0;
   		rt.type = 0;
+		rt.length = 0;
+		return rt;
 		}
   }
  }
@@ -680,7 +927,14 @@ struct arglist security_hole(globals, vars)
  char * data = arg_get_value(vars, "data");
  char * asc_port = arg_get_value(vars, "port");
 
- 
+ if((harg_get_int(globals, "standalone"))==1)
+ {
+  if(data)
+   fprintf(stderr, "%s\n", data);
+  else
+   fprintf(stderr, "Success\n");
+ }
+   
  if(!proto)proto = arg_get_value(vars, "protocol");
  
 #ifdef DEBUG
@@ -753,6 +1007,14 @@ struct arglist security_warning(globals, vars)
  char * data = arg_get_value(vars, "data");
  char * asc_port = arg_get_value(vars, "port");
  
+  if((harg_get_int(globals, "standalone"))==1)
+ {
+  if(data)
+   fprintf(stderr, "%s\n", data);
+  else
+   fprintf(stderr, "Success\n");
+ 
+ }
  
  if(!proto)proto = arg_get_value(vars, "protocol");
 #ifdef DEBUG_NASL
@@ -821,47 +1083,51 @@ struct arglist pkt_recv(globals, vars)
 #ifdef DEBUG_NASL
  rt.value = prompt(globals, "pkt_recv : ");
 #else
- FD_ZERO(&rd);
- FD_SET(soc, &rd);
+ if(arg_get_length(vars, "socket") != sizeof(int))
+ {
+  return rt;
+ }
+ 
  tv.tv_sec = harg_get_int(globals, "read_timeout");
  if(!tv.tv_sec)tv.tv_sec = 15;
  if(asc_to)tv.tv_sec = atoi(asc_to);
-#ifdef DEBUG
- printf("timeout : %ld\n", tv.tv_sec);
-#endif
- tv.tv_usec = 0;
- if(select(soc+1, &rd, NULL, NULL, &tv)>0)
- {
-  struct sockaddr_in * soca;
-  int soca_len = 0;
-  if(!(soca = harg_get_ptr(udp_sockets, asc)))
-   new_len = recv(soc, data, len, 0);
-  else
-   new_len = recvfrom(soc, data, len, 0, (struct sockaddr *)soca, &soca_len);
+ tv.tv_usec = 0; 
   
-  if(new_len <= 0)
+ for(;;)
+ {
+  FD_ZERO(&rd);
+  FD_SET(soc, &rd);
+  
+  if(select(soc+1, &rd, NULL, NULL, &tv)>0)
   {
-   bzero(&rt, sizeof(rt));
-   return rt;
+   struct sockaddr_in * soca;
+   int soca_len = 0;
+   int e;
+   if(!(soca = harg_get_ptr(udp_sockets, asc)))
+    e = recv(soc, data+new_len, len-new_len, 0);
+   else
+    e = recvfrom(soc, data+new_len, len-new_len, 0, (struct sockaddr *)soca, &soca_len);
+  
+   if(e <= 0)
+   {
+    if(!new_len)
+    {
+     bzero(&rt, sizeof(rt));
+     rt.type = VAR_STR;
+     return rt;
+    }
+    else break;
+   }
+   else new_len+=e;
+   if(new_len >= len)break;
   }
-  rt.value = (void*)nasl_malloc(globals,new_len+1);
-#ifdef DEBUG
-  printf("%s socket\n", soca?"udp":"tcp");
-  printf("Received %d bytes\n", new_len);
-  {
-   int k;
-   printf("Data : ");
-   for(k=0;k<new_len;k++)printf("0x%x ", data[k]);
-   printf("\n");
-  }
-#endif /* DEBUG */
+  else break;
+  
+  bzero(&tv, sizeof(tv));
+  tv.tv_usec = 100;
  }
- else {
- 	rt.value = NULL;
-	rt.length = 0;
-        rt.type = VAR_STR;
-	new_len = 0;
-       }
+ 
+ rt.value = (void*)nasl_malloc(globals,new_len+1);
 #endif
  nasl_free(globals, asc);
  if(new_len>0)
@@ -998,20 +1264,26 @@ struct arglist telnet_init(globals, vars)
 	return(rt);
 	}
  iac = 255;
+ 
+ tv.tv_sec = harg_get_int(globals, "read_timeout");
+ if(!tv.tv_sec)tv.tv_sec = 15;
+ tv.tv_usec = 0;
+  
+  
  while(iac == 255)
  {
+  int n = 0;
   FD_ZERO(&rd);
   FD_SET(soc, &rd);
-  tv.tv_sec = harg_get_int(globals, "read_timeout");
-  if(!tv.tv_sec)tv.tv_sec = 15;
-  tv.tv_usec = 0;
   select(soc+1, &rd, NULL, NULL, &tv);
   if(!FD_ISSET(soc, &rd))return(rt);
-  recv(soc, data, 3, 0);
-  if(iac!=255)break;
+  n = recv(soc, data, 3, 0);
+  if((iac!=255)||(n<=0))break;
   if((code == 251)||(code == 252))code = 254; /* WILL , WONT -> DON'T */
   else if((code == 253)||(code == 254))code = 252; /* DO,DONT -> WONT */
   send(soc, data,3,0);
+  bzero(&tv, sizeof(tv));
+  tv.tv_sec = 1;
  }
  FD_ZERO(&rd);
  FD_SET(soc, &rd);
