@@ -25,7 +25,50 @@
 #include "sanitize.h"
 #include "defines.h"
 #include "nasl_memory.h"
+#include "regex.h"
+#include "instruction.h"
 
+/*
+ * This function does nothing. It's is used to 
+ * store the calls to upcoming functions such as script_bugtraq_id()
+ */
+struct arglist 
+nasl_null_function()
+{ 
+ struct arglist rt;
+ bzero(&rt, sizeof(rt));
+ return rt;
+}
+
+
+/*
+ * Set the internal variable '___nasl_return' to 
+ * our argument
+ */
+struct arglist 
+nasl_return(globals, args)
+  harglst* globals;
+  struct arglist * args;
+{
+  struct arglist rt;
+  char * val;
+   
+  val = args->value;
+  if(!val)val = "0";
+  rt = sanitize_variable(globals, val);
+  if(rt.value)
+  {
+    affect_var(globals, rt, "__nasl_return");
+    bzero(&rt, sizeof(rt));
+    rt.type = PKT_ERROR(ERR_FUNC_EXIT);
+    return rt;
+  }
+  rt.type = PKT_ERROR(ERR_FUNC_EXIT);
+  return rt;
+}
+
+
+  
 struct arglist nasl_asctime(globals, args)
  harglst * globals;
  struct arglist * args;
@@ -229,7 +272,6 @@ struct arglist pkt_ord(globals, args)
   {
    char*c = ar.value;
    char * ret = nasl_malloc(globals,12);
-   printf("->%d\n", c[0]);
    sprintf(ret, "%u", (u_char)c[0]);
    rt.type = VAR_STR|STR_ALL_DIGIT;
    rt.length = strlen(ret);
@@ -336,7 +378,7 @@ struct arglist pkt_strstr(globals, args)
   if((sa.type & sb.type) & VAR_STR)
   {
    if(sa.length < sb.length)rt.value = NULL;
-   else rt.value = memmem(sa.value,sa.length, sb.value, sb.length);
+   else rt.value = (void*)memmem(sa.value,sa.length, sb.value, sb.length);
    rt.length = sa.length - ((int)rt.value - (int)sa.value);
    if(!rt.value){
    	rt.value = nasl_strdup(globals, "0");
@@ -349,6 +391,299 @@ struct arglist pkt_strstr(globals, args)
  }
  return(rt);
 }
+
+
+
+/*
+ * regex syntax :
+ *
+ *	ereg(pattern, string)
+ */
+struct arglist ereg(globals, args)
+ harglst * globals;
+ struct arglist * args;
+{
+ char * pattern = arg_get_value(args, "pattern");
+ char * string = arg_get_value(args, "string");
+ int length_string = arg_get_length(args, "string");
+ struct arglist rt;
+ rt.type = -1;
+ rt.value = NULL;
+ if(pattern && string)
+ {
+  char fastmap[1 << 8];
+  struct re_pattern_buffer pb;
+  struct re_registers regs;
+ 
+  pb.allocated = 8;
+  pb.buffer = nasl_malloc(globals, 8);
+  pb.fastmap = fastmap;
+  pb.translate = 0;
+  re_set_syntax(RE_SYNTAX_POSIX_EGREP);
+  re_compile_pattern(pattern, arg_get_length(args, "pattern"), &pb);
+  re_compile_fastmap(&pb);
+  if(!re_search(&pb, 
+  		string,
+		length_string,
+		0, 
+  		length_string,
+		&regs))
+  {
+    rt.type = VAR_STR;
+    rt.value = strdup("1");
+    rt.length = 1;
+  }
+  else
+  {
+    rt.length = 0;
+    rt.value = NULL;
+    rt.type = 0;
+  }
+  nasl_free(globals, pb.buffer);
+  }
+  /* if(sa.type & VAR_DELETE)free(sa.value); */
+  return(rt);
+}
+
+/*
+ * ereg_replace :
+ *
+ * result = ereg_replace(pattern, replace, string, [icase])
+ *
+ */
+ 
+ 
+#define NS 1024
+/*
+ * Ripped from the php3 team !
+ */ 
+/* this is the meat and potatoes of regex replacement! */
+static char * _regreplace(const char *pattern, 
+		const char *replace, const char *string, int icase, int extended)
+{
+	regex_t re;
+	regmatch_t subs[NS];
+
+	char *buf,	/* buf is where we build the replaced string */
+	     *nbuf,	/* nbuf is used when we grow the buffer */
+		 *walkbuf; /* used to walk buf when replacing backrefs */
+	const char *walk; /* used to walk replacement string for backrefs */
+	int buf_len;
+	int pos, tmp, string_len, new_l;
+	int err, copts = 0;
+
+	string_len = strlen(string);
+
+	if (icase)
+		copts = REG_ICASE;
+	if (extended)
+		copts |= REG_EXTENDED;
+	err = regcomp(&re, pattern, copts);
+	if (err) {
+		return NULL;
+	}
+
+	/* start with a buffer that is twice the size of the stringo
+	   we're doing replacements in */
+	buf_len = 2 * string_len + 1;
+	buf = emalloc(buf_len * sizeof(char));
+	if (!buf) {
+		return NULL;
+	}
+
+	err = pos = 0;
+	buf[0] = '\0';
+
+	while (!err) {
+		err = regexec(&re, &string[pos], (size_t) NS, subs, (pos ? REG_NOTBOL : 0));
+
+		if (err && err != REG_NOMATCH) {
+			return(NULL);
+		}
+		if (!err) {
+			/* backref replacement is done in two passes:
+			   1) find out how long the string will be, and allocate buf
+			   2) copy the part before match, replacement and backrefs to buf
+
+			   Jaakko Hyvtti <Jaakko.Hyvatti@iki.fi>
+			   */
+
+			new_l = strlen(buf) + subs[0].rm_so; /* part before the match */
+			walk = replace;
+			while (*walk)
+				if ('\\' == *walk
+					&& '0' <= walk[1] && '9' >= walk[1]
+					&& subs[walk[1] - '0'].rm_so > -1
+					&& subs[walk[1] - '0'].rm_eo > -1) {
+					new_l += subs[walk[1] - '0'].rm_eo
+						- subs[walk[1] - '0'].rm_so;
+					walk += 2;
+				} else {
+					new_l++;
+					walk++;
+				}
+
+			if (new_l + 1 > buf_len) {
+				buf_len = 1 + buf_len + 2 * new_l;
+				nbuf = emalloc(buf_len);
+				strcpy(nbuf, buf);
+				efree(&buf);
+				buf = nbuf;
+			}
+			tmp = strlen(buf);
+			/* copy the part of the string before the match */
+			strncat(buf, &string[pos], subs[0].rm_so);
+
+			/* copy replacement and backrefs */
+			walkbuf = &buf[tmp + subs[0].rm_so];
+			walk = replace;
+			while (*walk)
+				if ('\\' == *walk
+					&& '0' <= walk[1] && '9' >= walk[1]
+					&& subs[walk[1] - '0'].rm_so > -1
+					&& subs[walk[1] - '0'].rm_eo > -1) {
+					tmp = subs[walk[1] - '0'].rm_eo
+						- subs[walk[1] - '0'].rm_so;
+					memcpy (walkbuf,
+							&string[pos + subs[walk[1] - '0'].rm_so],
+							tmp);
+					walkbuf += tmp;
+					walk += 2;
+				} else
+					*walkbuf++ = *walk++;
+			*walkbuf = '\0';
+
+			/* and get ready to keep looking for replacements */
+			if (subs[0].rm_so == subs[0].rm_eo) {
+				if (subs[0].rm_so + pos >= string_len)
+					break;
+				new_l = strlen (buf) + 1;
+				if (new_l + 1 > buf_len) {
+					buf_len = 1 + buf_len + 2 * new_l;
+					nbuf = emalloc(buf_len * sizeof(char));
+					strcpy(nbuf, buf);
+					efree(&buf);
+					buf = nbuf;
+				}
+				pos += subs[0].rm_eo + 1;
+				buf [new_l-1] = string [pos-1];
+				buf [new_l] = '\0';
+			} else {
+				pos += subs[0].rm_eo;
+			}
+		} else { /* REG_NOMATCH */
+			new_l = strlen(buf) + strlen(&string[pos]);
+			if (new_l + 1 > buf_len) {
+				buf_len = new_l + 1; /* now we know exactly how long it is */
+				nbuf = emalloc(buf_len * sizeof(char));
+				strcpy(nbuf, buf);
+				efree(&buf);
+				buf = nbuf;
+			}
+			/* stick that last bit of string on our output */
+			strcat(buf, &string[pos]);
+			
+		}
+	}
+
+	buf [new_l] = '\0';
+
+	/* whew. */
+	return (buf);
+}
+
+
+struct arglist ereg_replace(globals, args)
+ harglst * globals;
+ struct arglist * args;
+{
+ char * pattern = arg_get_value(args, "pattern");
+ char * replace = arg_get_value(args, "replace");
+ char * string  = arg_get_value(args, "string");
+ char * icase   = arg_get_value(args, "icase");
+ struct arglist rt;
+ char * r; 
+ bzero(&rt, sizeof(rt));
+ 
+ if(!pattern || !replace || !string)
+ {
+  fprintf(stderr, "Usage : ereg_replace(string:<string>, pattern:<pat>, replace:<replace>, icase:<TRUE|FALSE>\n");
+  return(rt);
+ }
+
+ r = _regreplace(pattern, replace, string, icase?(icase[0]=='1'):0, 1);
+ rt.value  = nasl_strdup(globals, r);
+ rt.length = strlen(r);
+ rt.type = VAR_STR;
+ return(rt);
+}
+
+/*
+ * regex syntax :
+ *
+ *	egrep(pattern, string)
+ */
+struct arglist egrep(globals, args)
+ harglst * globals;
+ struct arglist * args;
+{
+ char * pattern = arg_get_value(args, "pattern");
+ char * string = arg_get_value(args, "string");
+ int length_string = arg_get_length(args, "string");
+ struct arglist rt;
+ rt.type = -1;
+ rt.value = NULL;
+ if(pattern && string)
+ {
+  char fastmap[1 << 8];
+  struct re_pattern_buffer pb;
+  struct re_registers regs;
+  char * s;
+ 
+  
+  s = string;
+  while(s)
+  {
+  pb.allocated = 8;
+  pb.buffer = nasl_malloc(globals, 8);
+  pb.fastmap = fastmap;
+  pb.translate = 0;
+  re_set_syntax(RE_SYNTAX_POSIX_EGREP);
+  re_compile_pattern(pattern, arg_get_length(args, "pattern"), &pb);
+  re_compile_fastmap(&pb);
+  length_string = strlen(s);
+  if(!re_search(&pb, 
+  		s,
+		length_string,
+		0, 
+  		length_string,
+		&regs))
+     {
+      char * t = strchr(s, '\n');
+      rt.type = VAR_STR;
+      if(t)t[0]='\0';
+      rt.value = strdup(s);
+      rt.length = strlen(s);
+      if(t)t[0]='\n';
+       nasl_free(globals, pb.buffer);  
+      return rt;
+    }
+  nasl_free(globals, pb.buffer);  
+  s = strchr(s+1, '\n');
+  if(s)s+=sizeof(char);
+  if(!s || !strlen(s))break;
+  }
+  
+ 
+  rt.length = 0;
+  rt.value = NULL;
+  rt.type = 0;
+  
+  }
+  /* if(sa.type & VAR_DELETE)free(sa.value); */
+  return(rt);
+}
+
 struct arglist
 this_host(globals, args)
  harglst * globals;
@@ -480,16 +815,30 @@ raw_string(globals, args)
    {
    for(i=0,j=0;i<len;i++)
    {
-   {
    if(val[i]=='\\'){
    		if(val[i+1]=='n'){str[j++]='\n';i++;}
 		else if(val[i+1]=='t'){str[j++]='\t';i++;}
 		else if(val[i+1]=='r'){str[j++]='\r';i++;}
-		else if(val[i+1]==val[i]){str[j++]=val[i];i++;}
+		else if(val[i+1]=='x' && isxdigit(val[i+2]) &&
+				         isxdigit(val[i+3])) {
+		 int x=0;
+                  if(isdigit(val[i+2])) {
+                    x=(val[i+2]-'0')*16;
+                  }else{
+                    x=(10+tolower(val[i+2])-'a')*16;
+                  }
+                  if(isdigit(val[i+3])) {
+                    x += val[i+3]-'0';
+                  }else{
+                    x += tolower(val[i+3])+10-'a';
+                  }
+                  str[j++]=x;
+                  i+=3;
+		 }			 
+		else if(val[i+1]=='\\'){str[j++]=val[i];i++;}
 		else i++;
 		}
    else str[j++] = val[i];
-   }
   }
  current_len = j;
  }
@@ -558,7 +907,23 @@ string(globals, args)
    		if(val[i+1]=='n'){str[j++]='\n';i++;}
 		else if(val[i+1]=='t'){str[j++]='\t';i++;}
 		else if(val[i+1]=='r'){str[j++]='\r';i++;}
-		else if(val[i+1]==val[i]){str[j++]=val[i];i++;}
+		else if(val[i+1]=='x' && isxdigit(val[i+2]) &&
+				         isxdigit(val[i+3])) {
+		 int x=0;
+                  if(isdigit(val[i+2])) {
+                    x=(val[i+2]-'0')*16;
+                  }else{
+                    x=(10+tolower(val[i+2])-'a')*16;
+                  }
+                  if(isdigit(val[i+3])) {
+                    x += val[i+3]-'0';
+                  }else{
+                    x += tolower(val[i+3])+10-'a';
+                  }
+                  str[j++]=x;
+                  i+=3;
+		 }
+		else if(val[i+1]=='\\'){str[j++]=val[i];i++;}
 		else i++;
 		}
    else {
@@ -648,23 +1013,51 @@ void
 nasl_exit(globals)
  harglst * globals;
 { 
+ if(!globals) 
+   return;
+ else
+ {  
  harglst * pcaps = harg_get_harg(globals, "pcaps");
+ harglst * variables = harg_get_harg(globals, "variables");
+ harglst * types = harg_get_harg(globals,"variables_types");
  hargwalk *hw = harg_walk_init(pcaps);
- 
-  if(!getuid())
+ char * key;
+
+ while((key = (char*)harg_walk_next(hw)))
  {
-  char * key;
-  while((key = harg_walk_next(hw)))
-  {
    pcap_t * p = harg_get_ptr(pcaps, key);
    if(p)pcap_close(p);
-  }
  }
+ 
+ 
+ /*
+  * Close the remaining sockets
+  */
+ /*harg_walk_stop(hw);*/ 
+ hw = harg_walk_init(variables);
+ while((key = (char*)harg_walk_next(hw)))
+  {
+   int s = harg_get_int(variables, key);
+   if(s > 0)
+    {
+     int type = harg_get_int(types, key);
+     if(type > 0)
+     {
+      if(!(type & VAR_CONST))
+      {
+       shutdown(s, 2);
+       close(s);
+      }
+     }
+    }
+  }
+ /*harg_walk_stop(hw); */ 
  nasl_memory_cleanup(globals);
  close((int)harg_get_int(globals, "socket"));
  /*exit(code);*/
  harg_remove(globals, "memory_manager");
  harg_close_all(globals);
+ }
 }
 
 
