@@ -22,4 +22,10 @@ struct arglist pkt_hex(harglst*, struct arglist*);
 struct arglist shift_left(harglst*, struct arglist*);
 struct arglist shift_right(harglst*, struct arglist*);
 struct arglist asc_time(harglst*, struct arglist*);
+struct arglist ereg(harglst*, struct arglist*);
+struct arglist ereg_replace(harglst*, struct arglist*);
+struct arglist egrep(harglst*, struct arglist*);
+struct arglist nasl_asctime(harglst*, struct arglist*);
+struct arglist nasl_return(harglst*, struct arglist*);
+struct arglist nasl_null_function(harglst*, struct arglist*);
 #endif
